@@ -4,14 +4,13 @@ Near-replica of example from the readme.
 
 from pathlib import Path
 
-from typeddfs import TypedDfs
+from typeddfs._entries import TypedDfs
 
 
 def run() -> None:
     """
     Runs an example usage of typeddfs.
     """
-
     # Build me a Key-Value-Note class!
     # noinspection PyPep8Naming
     KeyValue = TypedDfs.example()
